# RHGitaly protocol definitions

gRPC methods in service traits generated by `tonic-build` do not have blanket
implementations, but the `call` method of the corresponding server struct does return
the `UNIMPLEMENTED` gRPC Status code for unknown methods
(see "Generated code extracts below).

Because the purpose of RHGitaly is to provide only a handful of methods, it would be
extremely tedious to implement all the service methods only to return
`Status::unimplemented("Not in RHGitaly")`, as it would entail copying over long
signatures each time Gitaly grows a new method. This is much unwanted friction in
the experimental phase of the RHGitaly project, and probably unacceptable in regular
maintenance as well, although it has the compiler prevent protocol divergence.

Therefore, we provide stripped down protocol definitions, with only the service
methods that are actually implemented in RHGitaly. The drawback is that these
will have to be maintained.

In a later move me might either:

- provide a script to generate the stripped down protocol files and run it from
  `generate-stubs`
- extend `tonic-build` to add blanket implementations in the service traits, returning
  `Status::UNIMPLEMENTED`. See also `prost` documentation about
  [Services plugins](https://crates.io/crates/prost/0.11.8#services), but that may
  entail unacceptable forking / code duplication.

## Content of stripped down protocol files

We chose to remove unused method declarations only, keeping all `message` declarations:
this will simplify maintenance (whether manual or scripted) and also make it easier to
add further methods.

In the original changeset, we defined the three methods that the experimental phase
is about right away. Later on, a method will be added together with its implementation.

## Extracts

We chose `RepositoryService.RepositoryExists` as example because it is both very
simple and less confusing than anything from `ServerService`.

### Service trait

```rust
    #[async_trait]
    pub trait RepositoryService: Send + Sync + 'static {
        /// This comment is left unintentionally blank.
        async fn repository_exists(
            &self,
            request: tonic::Request<super::RepositoryExistsRequest>,
        ) -> Result<tonic::Response<super::RepositoryExistsResponse>, tonic::Status>;
        /// (...)
    }
```

### Server struct

```rust
    impl<T, B> tonic::codegen::Service<http::Request<B>> for RepositoryServiceServer<T>
    where
        T: RepositoryService,
        B: Body + Send + 'static,
        B::Error: Into<StdError> + Send + 'static,
    {
        type Response = http::Response<tonic::body::BoxBody>;
        type Error = std::convert::Infallible;
        type Future = BoxFuture<Self::Response, Self::Error>;
        // (...)

        fn call(&mut self, req: http::Request<B>) -> Self::Future {
            let inner = self.inner.clone();
            match req.uri().path() {
                "/gitaly.RepositoryService/RepositoryExists" => {
                    // (...)
                }
                // (...)
                _ => {
                    Box::pin(async move {
                        Ok(
                            http::Response::builder()
                                .status(200)
                                .header("grpc-status", "12")
                                .header("content-type", "application/grpc")
                                .body(empty_body())
                                .unwrap(),
                        )
                    })
                }
            }
        }
    }
```

This default "12" for the "grpc-status" header is the right one:

```python
import grpc
grpc.StatusCode.UNIMPLEMENTED
<StatusCode.UNIMPLEMENTED: (12, 'unimplemented')>
```


