// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::error::Error;
use std::net::SocketAddr;
use std::path::PathBuf;
use std::sync::Arc;
use tokio::fs::remove_file;
use tokio::net::UnixListener;
use tokio_stream::wrappers::UnixListenerStream;
use tonic::transport::Server;

use rhgitaly::config::Config;
use rhgitaly::service::commit::commit_server;
use rhgitaly::service::repository::repository_server;
use rhgitaly::service::server::server_server;
use tracing::{info, Level};
use tracing_subscriber::FmtSubscriber;

fn setup_tracing() {
    // a builder for `FmtSubscriber`.
    let subscriber = FmtSubscriber::builder()
        .with_max_level(Level::INFO)
        .finish();

    tracing::subscriber::set_global_default(subscriber).expect("setting default subscriber failed");
}

#[derive(Debug)]
#[allow(clippy::upper_case_acronyms)]
enum BindAddress {
    Unix(PathBuf),
    TCP(SocketAddr),
}

/// Parse the URL to listen to
///
/// There are two cases: `unix:/some/path/sock` and `tcp://hostname:port`, where
/// hostname can only be an IP address (v4 or v6).
///
/// This is a toy implementation anyway, we'll add a proper dependency to the `url` crate later on
fn parse_listen_url(url: &str) -> Result<BindAddress, Box<dyn Error>> {
    if let Some(path) = url.strip_prefix("unix:") {
        return Ok(BindAddress::Unix(path.into()));
    }
    if let Some(addr) = url.strip_prefix("tcp://") {
        return Ok(BindAddress::TCP(addr.parse()?));
    }
    Err("Unsupported URL".into())
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let config = Arc::new(Config::from_env());
    setup_tracing();

    let server = Server::builder()
        .add_service(server_server())
        .add_service(commit_server(&config))
        .add_service(repository_server(&config));

    let bind_addr = parse_listen_url(&config.listen_url)?;
    info!("RHGitaly starting, binding to {:?}", &bind_addr);

    match bind_addr {
        BindAddress::TCP(addr) => server.serve(addr).await?,
        BindAddress::Unix(ref path) => {
            let uds = UnixListener::bind(path)?;
            let uds_stream = UnixListenerStream::new(uds);
            server.serve_with_incoming(uds_stream).await?
        }
    };

    // TODO perform this cleanup in a SIGINT/SIGTERM handler (currently the only way to
    // stop RHGitaly)
    if let BindAddress::Unix(path) = bind_addr {
        info!("Removing Unix Domain socket file at '{}'", &path.display());
        remove_file(&path).await?
    }

    Ok(())
}
