// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::fmt;
use std::marker::Send;
use std::path::PathBuf;
use std::sync::Arc;

use tokio::sync::mpsc;
use tokio_stream::wrappers::ReceiverStream;
use tonic::{Response, Status};

use hg::config::Config as CoreConfig;
use hg::repo::{Repo, RepoError};

use super::config::Config;
use super::gitaly::Repository;
use super::streaming::BlockingResponseSender;

/// Represent errors that are due to a wrong repository specification.
///
/// In terms of gRPC methods, the specifications is usually enclosed in  a [`Repository`] message
/// and these errors are considered to be client errors.
#[derive(Debug, PartialEq, Eq)]
pub enum RepoSpecError {
    MissingSpecification,
    UnknownStorage(String),
    RepoNotFound(PathBuf),
}

/// Represent errors loading a repository (bad specification or internal errors)
#[derive(Debug, derive_more::From)] // TODO add PartialEq, but do it in core for RepoError first
pub enum RepoLoadError {
    #[from]
    SpecError(RepoSpecError),
    LoadError(RepoError),
}

impl From<RepoError> for RepoLoadError {
    fn from(value: RepoError) -> Self {
        if let RepoError::NotFound { at } = value {
            return Self::SpecError(RepoSpecError::RepoNotFound(at));
        }
        Self::LoadError(value)
    }
}

/// Default conversion of ['RepoSpecError'] into a gRPC ['Status']
///
/// This function does not care to precisely match the error details, focusing on the error
/// codes instead.
///
/// The resulting codes match the most common behaviour of Gitaly, which actually behaves more
/// and more like this with time (e.g., as internal Git error get catched and handled).
pub fn default_repo_spec_error_status(err: RepoSpecError) -> Status {
    match err {
        RepoSpecError::MissingSpecification => {
            Status::invalid_argument("missing repository specification")
        }
        RepoSpecError::UnknownStorage(storage) => Status::invalid_argument(format!(
            "GetStorageByName: no such storage: \"{}\"",
            storage
        )),
        RepoSpecError::RepoNotFound(at) => {
            Status::not_found(format!("repository at \"{}\" not found", at.display()))
        }
    }
}

pub fn repo_path(config: &Config, repo: &Repository) -> Result<PathBuf, RepoSpecError> {
    if repo.storage_name != "default" {
        return Err(RepoSpecError::UnknownStorage(repo.storage_name.clone()));
    }
    let root = &config.repositories_root;

    let res = match repo.relative_path.strip_suffix(".git") {
        Some(stripped) => root.join(&(stripped.to_owned() + ".hg")),
        None => root.join(&repo.relative_path),
    };

    // TODO forbid climbing up (same in Python, will be necessary for HGitaly3, even
    // though clients are deeply trusted.
    Ok(res)
}

pub fn load_repo(config: &Config, opt_repo: Option<&Repository>) -> Result<Repo, RepoLoadError> {
    let repo = opt_repo.ok_or(RepoSpecError::MissingSpecification)?;
    // TODO load core config once and for all and store it in our Config.
    // but do we really want to have all this unrelated stuff around?
    let core_config =
        CoreConfig::load_non_repo().expect("Should have been able to read Mercurial config");
    // TODO better to use Repo::new_at_path, but it is private
    // (Repo::find does verifications that are useless to us.
    // At least it does not try to climb up when passed an explicit path)
    Ok(Repo::find(&core_config, Some(repo_path(config, repo)?))?)
}

/// Trait for requests with a repository field
///
/// It provides the needed uniformity for methods such as [`load_repo_and_stream`]
pub trait RequestWithRepo: Send + 'static {
    /// Grab a reference to the [`Repository`] field from the request.
    ///
    /// Like all submessages, the repository is optional.
    fn repository_ref(&self) -> Option<&Repository>;
}

/// Load a repository and initiate streaming responses
///
/// This setups the `mpsc` channel expected by Tonic and spawns a blocking task (typically run
/// in a separate thread) loads the repository, and passes over the repository and the transmitting
/// end of the channel to the caller supplied closure.
///
/// The `and_then` closure must perform its streaming by sending `Result<Resp, Status>` values
/// on the channel, using the provided [`BlockingResponseSender`].
///
/// If the repository loading fails, the appropriate gRPC error response is sent over
/// or logged if sending is not possible.
///
/// Because Gitaly's error responses are not uniform, and we want to match them closely,
/// ethe caller has to supply a callable for conversion of [`RepoSpecError`] to the appropriate
/// [`Status`]. The [`default_repo_spec_error_status`] function can be used in the majority of
/// cases and serve as an example for other cases.
pub fn load_repo_and_stream<Req: RequestWithRepo, Resp: fmt::Debug + Send + 'static>(
    config: Arc<Config>,
    request: Req,
    repo_spec_error_status: impl Fn(RepoSpecError) -> Status + Send + 'static,
    and_then: impl FnOnce(Req, Repo, BlockingResponseSender<Resp>) + Send + 'static,
) -> Result<Response<ReceiverStream<Result<Resp, Status>>>, Status> {
    // no point having channel capacity for several messages, since `blocking_send` is the only
    // way to use it.
    let (tx, rx) = mpsc::channel(1);
    tokio::task::spawn_blocking(move || {
        let btx: BlockingResponseSender<Resp> = tx.into();
        match load_repo(&config, request.repository_ref()) {
            Err(RepoLoadError::SpecError(e)) => btx.send(Err(repo_spec_error_status(e))),
            Err(RepoLoadError::LoadError(e)) => btx.send(Err(Status::internal(format!(
                "Error loading repository: {:?}",
                e
            )))),
            Ok(repo) => and_then(request, repo, btx),
        }
    });
    Ok(Response::new(ReceiverStream::new(rx)))
}

#[cfg(test)]
mod tests {

    use super::*;

    #[test]
    fn test_repo_path() {
        let config = Config {
            repositories_root: "/repos".into(),
            listen_url: "".into(),
        };
        let mut repo = Repository::default();
        repo.storage_name = "default".into();
        repo.relative_path = "foo/bar".into();
        assert_eq!(repo_path(&config, &repo), Ok("/repos/foo/bar".into()));

        repo.relative_path = "foo/bar.git".into();
        assert_eq!(repo_path(&config, &repo), Ok("/repos/foo/bar.hg".into()));
    }
}
