use std::sync::Arc;
use tokio_stream::wrappers::ReceiverStream;
use tonic::{Request, Response, Status};
use tracing::{info, instrument};

use hg::changelog::Changelog;
use hg::node::NodePrefix;
use hg::revlog::RevlogError;

use crate::config::Config;
use crate::gitaly::commit_service_server::{CommitService, CommitServiceServer};
use crate::gitaly::{ListCommitsByOidRequest, ListCommitsByOidResponse, Repository};
use crate::message;
use crate::repository::{default_repo_spec_error_status, load_repo_and_stream, RequestWithRepo};

#[derive(Debug)]
pub struct CommitServiceImpl {
    config: Arc<Config>,
}

const COMMITS_CHUNK_SIZE: usize = 50;

/// Return the [`ListCommitsByOidResponse`] for a chunk of oids (changeset ids)
///
/// Behaving like Gitaly imposes that
/// - oids that cannot be found in the repository are ignored. This is actually one of the
///   use cases for the `ListCommitsByOid` call: it can be called indiscriminately with anything
///   that looks like a hash, in particular from Markdown renderings
/// - oids that are not even hexadecimal are ignored, too. This is possibly coincidental.
///
/// [`RevlogError`] is returned in cases of repository corruption.
fn commits_by_oid_chunk(
    cl: &Changelog,
    oids: &[String],
) -> Result<ListCommitsByOidResponse, RevlogError> {
    // Not using Vec::with_capacity(oids.len()), because clients will often call this with a single
    // argument, precisely to know whether it is a commit id prefix in this repo (happens a lot
    // in Markdown rendering), and this will result in needless allocations.
    let mut commits = Vec::new();
    for oid in oids {
        match NodePrefix::from_hex(oid) {
            Err(_) => {
                continue; // non-hexadecimal values for oid are just ignored by Gitaly
            }
            Ok(node_pref) => match cl.rev_from_node(node_pref) {
                Ok(rev) => commits.push(message::commit(cl.entry_for_rev(rev)?)?),
                Err(RevlogError::InvalidRevision) => (),
                Err(e) => {
                    return Err(e);
                }
            },
        }
    }
    Ok(ListCommitsByOidResponse { commits })
}

impl RequestWithRepo for ListCommitsByOidRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

#[tonic::async_trait]
impl CommitService for CommitServiceImpl {
    type ListCommitsByOidStream = ReceiverStream<Result<ListCommitsByOidResponse, Status>>;

    #[instrument]
    async fn list_commits_by_oid(
        &self,
        request: Request<ListCommitsByOidRequest>,
    ) -> Result<Response<Self::ListCommitsByOidStream>, Status> {
        info!("Processing");

        let inner = request.into_inner();
        load_repo_and_stream(
            self.config.clone(),
            inner,
            default_repo_spec_error_status,
            move |req, repo, tx| match repo.changelog() {
                Ok(cl) => {
                    for oid_chunk in req.oid.chunks(COMMITS_CHUNK_SIZE) {
                        match commits_by_oid_chunk(&cl, oid_chunk) {
                            Ok(resp) => {
                                if !resp.commits.is_empty() {
                                    tx.send(Ok(resp));
                                }
                            }
                            Err(e) => tx.send(Err(Status::internal(format!(
                                "Repository corruption {:?}",
                                e
                            )))),
                        }
                    }
                }
                Err(e) => tx.send(Err(Status::internal(format!(
                    "Could not open repository changelog: {:?}",
                    e
                )))),
            },
        )
    }
}

/// Takes care of boilerplate that would instead be in the startup sequence.
pub fn commit_server(config: &Arc<Config>) -> CommitServiceServer<CommitServiceImpl> {
    CommitServiceServer::new(CommitServiceImpl {
        config: config.clone(),
    })
}
