// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use tokio::sync::mpsc::Sender;
use tonic::Status;
use tracing::error;

/// A specialization of [`Sender`] allowing only blocking sends and logging sending errors
///
/// Server streaming methods involving a repository typically spawn a new thread with
/// [`tokio::task::spawn_blocking`] because the primitives in `hg-core` are blocking. From this
/// thread, the only possible output is to use the [`Sender`] end of a multiple-producer,
/// single-consumer (mpsc) channel, with its `blocking_send` method.
///
/// This is currently the recommended way to send from synchronous to asynchronous code.
/// The only error condition to handle on these sends happens if the receiving side (which takes
/// care of sending the response to the client over the wire) is closing down or closed (e.g. for
/// loss of connectivity?). Since the channel is the only way for the task to report back
/// anything to the client, the only thing that can be done about such errors is to log them.
///
/// This struct also has the advantage to enforce that code from Tokio blocking tasks should never
/// use the (async) `send` method of the underlying mpsc Sender.
#[derive(Debug, derive_more::From)]
pub struct BlockingResponseSender<Resp>(Sender<Result<Resp, Status>>);

impl<Resp> BlockingResponseSender<Resp> {
    pub fn send(&self, res: Result<Resp, Status>) {
        self.0.blocking_send(res).unwrap_or_else(|send_error| {
            let msg = match send_error.0 {
                Ok(_) => "an ordinary response".into(),
                Err(status) => format!("an error response: {}", status),
            };
            // logging at error level because when we implement graceful
            // shutdown, we will try and avoid to write on channels that are
            // closing down.
            // TODO take the opportunity to fully log the original error
            error!("Channel closing down or closed while streaming {}", msg)
        });
    }
}
